/***********************************************************
 * $Id$
 *
 * FancyLibs
 *
 * Copyright (C) 2009-2011 ev-i GmbH & IBCL e.U.
 * Copyright (C) 2011-2013 ITEG IT-Engineers GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ************************************************************/

/****************************************************************************
 **
 ** Author: Christoph Lechleitner / ITEG
 **
 ** Interface for wrappers around wvdial or the like.
 **
 *****************************************************************************/

#ifndef FANCYSOCK_DIALER_H
#define FANCYSOCK_DIALER_H

#define FANCYSOCK_DIALER_BEAN "Fancy::Sock::Dialer"

#include <fancysock.h>

#include <fancythread.h>

//#include <string>

namespace Fancy {
class ConfigOptions;
class List;
//class Map;

namespace Sock {

class Address;


/**
 * A dialer callback receives connectivity status updates from a Dialer (see below).
 */
class FANCYSOCK_API DialerStatusCallback : virtual public Fancy::Object
{
public:
    typedef Fancy::ObjPtr<DialerStatusCallback> ptr_t;

    /**
     * Creates a simple instance just logging callbacks.
     */
    static ptr_t newSimpleInstance();

    /**
     * Called after the connection came up.
     */
    virtual void onDialConnectionCreated() = 0;

    /**
     * Called after the dial attempt has been given up.
     * The dialer automatically makes another attempt unless disconnect() has or is been called before the breath-taking 5 seconds are over.
     * It should be OK to call disconnect() in this handler.
     *
     * nrFailedAttempts contains the number of consecutive attempt failures.
     */
    virtual void onDialAttemptFailed(int nrFailedDialAttempts) = 0;

    /**
     * Called after the connection died.
     */
    virtual void onDialConnectionLost() = 0;

    /**
     * 'asks' if everything is ready to dial.
     * This allows to
     * - check back with user within an application
     * - switch on hardware components
     * - prepare monitoring processes
     * ...
     *
     * It is called _before_ an eventual PREPARE-MODEM-SCRIPT
     *
     * If false is returned, no dial attempt is made
     */
    virtual bool checkOKToDial() = 0;

protected:

};


/**
 * A dialer uses some WAN connection technology provided by the OS.
 * For now only Linux' wvdial is supported.
 *
 * Most dialers will set the program locale (resp. LC_ALL) to "C", to make subprogram outputs more predictable.
 */
class FANCYSOCK_API Dialer : virtual public Fancy::Runnable
{
public:
    typedef Fancy::ObjPtr<Dialer> ptr_t;

#ifndef WIN32

#define FANCYWVDIALER_OPTION_CONFIGFILE    "CONFIGFILE"
#define FANCYWVDIALER_OPTION_SECTION       "SECTION"
#define FANCYWVDIALER_OPTION_PREPARESCRIPT "PREPARE-MODEM-SCRIPT"
#define FANCYWVDIALER_OPTION_CLEANUPSCRIPT "CLEANUP-MODEM-SCRIPT"

#define FANCYWVDIALER_OPTION_CONNECT_TIMEOUT_SECS    "CONNECT-TIMEOUT-SECS"
#define FANCYWVDIALER_OPTION_DISCONNECT_TIMEOUT_SECS "DISCONNECT-TIMEOUT-SECS"

    /**
     * Instantiates a dialer facilitating wvdial.
     * wvidal must be configured properly, i.e. /etc/wvdial.conf must be valid.
     * Unfortunate situations like modems showing up at alternating devices (ttyUSB0/ttyUSB3) are without the scope of this implementation.
     *
     * Options supported by WVDialer instances are:
     *
     * "CONFIGFILE" (FANCYWVDIALER_OPTION_CONFIGFILE):
     *   Custom wvdial.conf, i.e. "/etc/myapp/mywvial.conf".
     *   Default: /etc/wvdial.conf
     *
     * "SECTION" (FANCYWVDIALER_OPTION_SECTION):
     *   Custom wvdial section name, i.i "mySectionName"
     *   Default: Empty, i.e. no section nameadded to wvdial call
     *
     * "PREPARE-MODEM-SCRIPT" (FANCYWVDIALER_OPTION_PREPARESCRIPT):
     *   Optional, full path to a script to call to prepare the modem, i.e. power it up.
     *   Called before wvdial.
     *   If the script returns a non-0 exit code, the dial attempt is aborted, and an eventual CLEANUP-MODEM-SCRIPT is called.
     *   To ignore errors from an underlying tool, end the script with "exit 0".
     *   Default: Empty, i.e. no preparation actions required.
     *
     * "CLEANUP-MODEM-SCRIPT" (FANCYWVDIALER_OPTION_CLEANUPSCRIPT):
     *   Optional, full path to a script to call to cleanup the modem, i.e. power it down.
     *   Called after killing the wvdial process.
     *   The script's return code may be logged, but is ignored otherwise.
     *   Default: Empty, i.e. no cleanup actions required.
     *
     * "CONNECT-TIMEOUT-SECS" (FANCYWVDIALER_OPTION_CONNECT_TIMEOUT_SECS)
     *   Optional. Defaults to 60 seconds.
     *   If no connection is created within this time, it's given up and onDialAttemptFailed() is called.
     *   Another attempt is made after taking a breath.
     *
     * "DISCONNECT-TIMEOUT-SECS" (FANCYWVDIALER_OPTION_DISCONNECT_TIMEOUT_SECS)
     *
     */
    static ptr_t newWVDialInstance();

    /**
     * Checks wether wvdial environment is avaiable.
     * This is satisfied if:
     * - /usr/bin/wvdial is present and executable
     * - /etc/wvdial.conf is present and readable
     */
    static bool isWVDialAvailable();

#endif

#define FANCYFAKEDIALER_OPTION_CONNECTDELAYMS    "CONNECT-DELAY-MILLIS"
#define FANCYFAKEDIALER_OPTION_DISCONNECTDELAYMS "DISCONNECT-DELAY-MILLIS"
#define FANCYFAKEDIALER_OPTION_WANUPFILE         "WAN-UP-FILE"

    /**
     * Instantiates a dialer facilitating a NOP- or pseudo-dialer.
     *
     * Options supported by Fake instance are:
     *
     * "CONNECT-DELAY-MILLIS" (FANCYFAKEDIALER_OPTION_CONNECTDELAYMS):
     *   Artifical delay between connect() call and successful connection in ms.
     *   Default: 2500 ms.
     *   Negative value: Connection attempt will fail.
     *
     * "DISCONNECT-DELAY-MILLIS" (FANCYFAKEDIALER_OPTION_DISCONNECTDELAYMS):
     *   Artifical delay between disconnect() call and successful disconnection in ms.
     *   Default: 1000 ms.
     *   Negative value: No automatic disconnection.
     *
     * "WAN-UP-FILE" (FANCYFAKEDIALER_OPTION_WANUPFILE):
     *   Name of file which's existence means that the pseudo connection is up.
     *   If [DIS]CONNECT-DELAY-MILLIS is >= 0, it is automatically created/removed.
     *   Otherwise it's existence is only monitored.
     *   Default:
     *     /tmp/fancysock-fakedialer-wan-up.txt
     *   resp.
     *     %TEMP%\fancysock-fakedialer-wan-up.txt
     */
    static ptr_t newFakeInstance();

    /**
     * Tells the dialer to connect. May be called before start resp. run.
     * Expect the actual dialing process to need between 1 and 150 seconds.
     */
    virtual void connect() = 0;

    /**
     * Tells the dialer to disconnect.
     * Expect the actual disconnection process to need between 1 and 30 seconds.
     */
    virtual void disconnect() = 0;

    /**
     * Tells the dialer to disconnect, and re-connect right away.
     * Expect the actual re-connection process to need between 1 and 180 seconds.
     */
    virtual void reconnect() = 0;

    /**
     * Tells the dialer to re-check it's connectivity ASAP.
     */
    virtual void recheckConnection() = 0;

    /**
     * Returns wether the connection is _considered_ to be up.
     * Does not guarantee real connectivity.
     */
    virtual bool isConnectionConsideredUp() = 0;

    /**
     * Must be called before connect() call.
     * Supported options should be documented at the ac. newFooVariantInstance() call.
     */
    virtual void setConnectionOptions(
    		const Fancy::ObjPtr<Fancy::ConfigOptions> &options) = 0;

    /**
     * Must be called before starting the Dialer.
     * Sets the interval of connectivity checks (while connection considered up), and optional ping parameters.
     * Technology dependent checks are always used first, if available.
     * If this does not failed and pingAddress is set non-null, a ping is performed.
     * Setting checkIntervalSeconds <= 0 disables regular checks
     */
    virtual void setCheckOptions(
            unsigned checkIntervalSeconds = 30,
            const Fancy::ObjPtr<Address> &pingAddress = 0,
            int pingMaxHops = 63,
            int pingTimeoutMillis = 2000,
            int pingNrPings = 1,
            int pingAttempts = 3) = 0;

    /**
     * Sets a callback that receives status updates
     */
    virtual void setStatusCallback(
            const Fancy::ObjPtr<DialerStatusCallback> &statusCallback) = 0;

protected:
    Dialer();
    virtual ~Dialer();

};

}} // end namespace Fancy::Sock

#endif // ! FANCYSOCK_DIALER_H
