/***********************************************************
 * $Id$
 *
 * FancyLibs
 *
 * Copyright (C) 2014 ITEG IT-Engineers GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ************************************************************/

#ifndef FANCYSOCKHTTPCLIENT_H_
#define FANCYSOCKHTTPCLIENT_H_

#include <fancysock.h>

#include <fancyrunnable.h>

namespace Fancy {

class Map;
class URL;

namespace Sock {

class DeviceFactory;
class HttpTransport;
class ResolveOptions;

/**
 * A general-purpose HTTP client, which may be used as a singleton in a
 * multi-threaded environment.
 *
 * If you start an HttpClient instance in a @c Thread or @c ThreadPool,
 * the garbage collector is activated and connections will be reused and
 * closed after the configured @c keepConnectionTimeout.
 */
class FANCYSOCK_API HttpClient: virtual public Fancy::Runnable {

public:
    typedef Fancy::ObjPtr<HttpClient> ptr_t;

    /**
     * Instantiate an HTTP client instance.
     */
    static ptr_t newInstance();

    /**
     * Issue a GET request to the given URL with the given request headers.
     *
     * The methods to be called on the returned transport are
     *
     * <pre>
     *   transport = client->doGet(url,headers);
     *   headers = transport->openReadDoc(30000);
     *   while (transport->read(...) > 0) {
     *       // store data...
     *   }
     * </pre>
     *
     * Calling <code>transport->close()</code> on the returned transport
     * explicitly closes the socket and inhibits the reuse of the socket.
     *
     * Normally the transport is put back to the client for later reuse,
     * when <code>transport->read()</code> returns <code>0</code>.
     */
    virtual Fancy::ObjPtr<HttpTransport> doGet(
            const Fancy::ObjPtr<Fancy::URL> &url,
            const Fancy::ObjPtr<Fancy::Map> &requestHeaders) = 0;

    /**
     * Issue a POST request to the given URL with the given request headers
     * and form fields in <code>application/x-www-form-urlencoded</code>.
     *
     * With non-null formData the request header field
     * "content-type" is automatically set to "application/x-www-form-urlencoded",
     * do not try to set a "Content-Type"!
     *
     * The methods to be called on the returned transport are
     *
     * <pre>
     *   transport = client->doPost(url,formData,headers);
     *   headers = transport->openReadDoc(30000);
     *   while (transport->read(...) > 0) {
     *       // store data...
     *   }
     * </pre>
     *
     * Calling <code>transport->close()</code> on the returned transport
     * explicitly closes the socket and inhibits the reuse of the socket.
     *
     * Normally the transport is put back to the client for later reuse,
     * when <code>transport->read()</code> returns <code>0</code>.
     */
    virtual Fancy::ObjPtr<HttpTransport> doPost(
            const Fancy::ObjPtr<Fancy::URL> &url,
            const Fancy::ObjPtr<Fancy::Map> &formData,
            const Fancy::ObjPtr<Fancy::Map> &requestHeaders) = 0;

    /**
     * Issue a POST request to the given URL with the given request headers
     * and the request data written by the user to the returned
     * transport.
     *
     * The methods to be called on the returned transport are
     *
     * <pre>
     *   transport = client->doPost(url,headers);
     *
     *   // write
     *   transport->write(...);
     *   transport->flush();
     *
     *   headers = transport->openReadDoc(30000);
     *   while (transport->read(...) > 0) {
     *       // store data...
     *   }
     * </pre>
     *
     * Calling <code>transport->close()</code> on the returned transport
     * explicitly closes the socket and inhibits the reuse of the socket.
     *
     * Normally the transport is put back to the client for later reuse,
     * when <code>transport->read()</code> returns <code>0</code>.
     */
    virtual Fancy::ObjPtr<HttpTransport> doPost(
            const Fancy::ObjPtr<Fancy::URL> &url,
            const Fancy::ObjPtr<Fancy::Map> &requestHeaders) = 0;

    /**
     * Issue a request with a custom method like PUT, HEAD, DELETE...
     * to the given URL with the given request headers
     * and the request data written
     *
     * The methods to be called on the returned transport are
     *
     * <pre>
     *   transport = client->doMethod("PUT",url,formData,headers);
     *
     *   // write
     *   transport->write(...);
     *   transport->flush();
     *
     *   headers = transport->openReadDoc(30000);
     *   while (transport->read(...) > 0) {
     *       // store data...
     *   }
     * </pre>
     *
     * If the request method does not require to send a body,
     * there's no need to write() and flush() the request body.
     *
     * Calling <code>transport->close()</code> on the returned transport
     * explicitly closes the socket and inhibits the reuse of the socket.
     *
     * Normally the transport is put back to the client for later reuse,
     * when <code>transport->read()</code> returns <code>0</code>.
     */
    virtual Fancy::ObjPtr<HttpTransport> doMethod(
            const std::string &method,
            const Fancy::ObjPtr<Fancy::URL> &url,
            const Fancy::ObjPtr<Fancy::Map> &requestHeaders) = 0;

    /**
     * Set, whether the client adds <code>Accept-Encoding: deflate</code>
     * to the request headers and transparently decodes the
     * compressed response stream.
     *
     * If you set the <code>Accept-Encoding</code> request header yourself,
     * the response stream is not transparently uncompressed.
     *
     * Default value is @c true.
     */
    virtual void setAcceptDeflate(bool b) = 0;

    /**
     * Return, whether deflate compression is configured.
     */
    virtual bool isAcceptDeflate() const = 0;

    /**
     * Return, whether this client handles cookie on it's own.
     * If @c true, the client maintains a cookie store to send
     * cookies back to the server, when a cookie was received in
     * advance by a <code>Set-Cookie</code> header.
     *
     * Default value is @c false.
     */
    virtual void setHandleCookies(bool b) = 0;

    /**
     * Return, whether this client handles cookie on it's own.
     */
    virtual bool isHandleCookies() = 0;

    /**
     * Set the device factory to use for <code>https</code> URLs.
     * If you want a connection timeout, please configure the
     * soRecvTimeout of the passed device factory.
     */
    virtual void setTlsDeviceFactory(const Fancy::ObjPtr<DeviceFactory> &tlsDeviceFactory) = 0;

    /**
     * Return the device factory configured for <code>https</code> URLs.
     */
    virtual Fancy::ObjPtr<DeviceFactory> getTlsDeviceFactory() const = 0;

    /**
     * Set the DNS resolve options to use.
     */
    virtual void setResolveOptions(const Fancy::ObjPtr<ResolveOptions> &resolveOptions) = 0;

    /**
     * Return the configured DNS resolve options.
     */
    virtual Fancy::ObjPtr<ResolveOptions> getResolveOptions() const = 0;

    /**
     * Set the SO_RECV_TIMEOUT of the connected sockets in milliseconds.
     * This timeout is used for the physical device connection unless
     * a TLS device factory has been configured.
     *
     * The default value is <code>30000</code>.
     */
    virtual void setSoTimeout(int millis) = 0;

    /**
     * Get the configured SO_RECV_TIMEOUT for the connected sockets in milliseconds.
     *
     * The default value is <code>30000</code>.
     */
    virtual int getSoTimeout() const = 0;

    /**
     * @param keepConnectionTimeout The timeout for keeping connections
     *                after a request, if this instance is started in a
     *                @c Thread or @c ThreadPool. 5000 milliseconds is
     *                the default value. If the server responded with a
     *                <code>Keep-Connection: max-age=x</code> header,
     *                the smaller value of this setting and the server value
     *                is chosen for an individual connection.
     */
    virtual void setKeepConnectionTimeout(int keepConnectionTimeout) = 0;

    /**
     * Get the configured timeout in millisecond for keeping connections
     */
    virtual int getKeepConnectionTimeout() const = 0;

protected:
    HttpClient();
    virtual ~HttpClient();
};

} /* namespace Sock */
} /* namespace Fancy */

#endif /* FANCYSOCKHTTPCLIENT_H_ */
