/***********************************************************
 * $Id$
 *
 * FancyLibs
 *
 * Copyright (C) 2014 ITEG IT-Engineers GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ************************************************************/

#ifndef FANCYSOCKHTTPCOOKIE_H_
#define FANCYSOCKHTTPCOOKIE_H_

#include <fancysock.h>

#include <fancyobject.h>
#include <string>

namespace Fancy {

class String;
class UtcTimestamp;

namespace Sock {

/**
 * An HTTP cookie as per <a href="http://tools.ietf.org/html/rfc6265">RFC 6265</a>.
 */
class FANCYSOCK_API HttpCookie: virtual public Fancy::Object {

    const std::string key;
    const std::string value;
    const Fancy::ObjPtr<Fancy::UtcTimestamp> expires;
    const Fancy::ObjPtr<Fancy::String> domain;
    const Fancy::ObjPtr<Fancy::String> path;
    const bool secure;
    const bool httpOnly;

public:
    typedef Fancy::ObjPtr<HttpCookie> ptr_t;

    /**
     * Instantiate a cookie from an HTTP set-cookie header value.
     *
     * If the @c setCookie line does not contain a path, use the path of
     * the originating request URI for referencing this cookie as per
     * <a href="http://tools.ietf.org/html/rfc6265#section-5.1.4">Section 5.1.4 of RFC6265</a>
     */
    static ptr_t newInstance(const std::string &setCookie);

    /**
     * Instantiate a cookie for given key and value and
     * optional expires, path and domain.
     */
    static ptr_t newInstance(
            const std::string &key,
            const std::string &value,
            const Fancy::ObjPtr<Fancy::UtcTimestamp> &expires,
            const Fancy::ObjPtr<Fancy::String> &domain,
            const Fancy::ObjPtr<Fancy::String> &path,
            bool secure,
            bool httpOnly
    );

    const std::string &getKey() const                            { return this->key; }
    const std::string &getValue() const                          { return this->value; }
    const Fancy::ObjPtr<Fancy::UtcTimestamp> &getExpires() const { return this->expires; }
    const Fancy::ObjPtr<Fancy::String> &getDomain() const        { return this->domain; }
    const Fancy::ObjPtr<Fancy::String> &getPath() const          { return this->path; }
    bool isSecure() const                                        { return this->secure; }
    bool isHttpOnly() const                                      { return this->httpOnly; }

    /**
     * Return A Cookie header value in the form <code>key=value</code> for
     *        sending it from the client to the server.
     */
    std::string toCookieValue() const;

    /**
     * Return a Set-Cookie header value containing the whole information
     * of the cookie.
     */
    std::string toString() const;

protected:
    HttpCookie(
            const std::string &key,
            const std::string &value,
            const Fancy::ObjPtr<Fancy::UtcTimestamp> &expires,
            const Fancy::ObjPtr<Fancy::String> &domain,
            const Fancy::ObjPtr<Fancy::String> &path,
            bool secure,
            bool httpOnly);

    virtual ~HttpCookie();
};

} /* namespace Sock */
} /* namespace Fancy */

#endif /* FANCYSOCKHTTPCLIENT_H_ */
