/***********************************************************
 * $Id$
 *
 * FancySock HTTPD helpers
 *
 * Copyright (C) 2013 ITEG IT-Engineers GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ************************************************************/

#ifndef FANCYSOCK_HTTPD_REQUEST_HANDLER_H_
#define FANCYSOCK_HTTPD_REQUEST_HANDLER_H_

#include <fancysock.h>

//#include <fancymap.h>
#include <fancyrunnable.h>

#include <string>


#define FANCYSOCK_HTTPD_CONTENTTYPE_HTML      "text/html"
#define FANCYSOCK_HTTPD_CONTENTTYPE_HTML_UTF8 "text/html; charset=utf-8"
#define FANCYSOCK_HTTPD_CONTENTTYPE_TEXT      "text/plain"
#define FANCYSOCK_HTTPD_CONTENTTYPE_TEXT_UTF8 "text/plain; charset=utf-8"
#define FANCYSOCK_HTTPD_CONTENTTYPE_XHTML     "application/xhtml+xml"
#define FANCYSOCK_HTTPD_CONTENTTYPE_XML       "text/xml"


namespace Fancy {
class Channel;
class DirEntry;
class List;
class Map;

namespace Sock {
class HttpServerTransport;

namespace HttpD {

/**
 * HttpD::RequestHandler provides a basic handler for connections to a HttpServerTransport.
 * It distributes the content work to a HttpDRequestHandler dependend on the URL path.
 *
 * Known implementations include:
 *   Fancy::Sock::ResourceProvider in this fancysock library, to deliver static content from the file system, see below.
 *   Fancy::HTTPd::EchoRequestHandler in the demo project fancyhttpd, a nice template for other HTTP server applications.
 *
 * We will soon provide a basic upload handler.
 */
class FANCYSOCK_API HttpDRequestHandler : public virtual Fancy::Object {

protected:
	HttpDRequestHandler();
	virtual ~HttpDRequestHandler();
public:

	typedef Fancy::ObjPtr<HttpDRequestHandler> ptr_t;

	typedef Fancy::ObjPtr<Fancy::Map> RequestHandlerMap;

	// no newInstance, for now this is a pure virtual class

	/**
	 * Main entry point for HTTP requests
	 */
	virtual void handleRequest(
			const Fancy::ObjPtr<Fancy::Map> & requestHeaders,
			const Fancy::ObjPtr<Fancy::Sock::HttpServerTransport> & httpServerTransport) = 0;


	/**
	 * Writes a fancylog log line that looks roughly like an access.log line from Apache
	 */
	virtual void writeAccessLog(
			const Fancy::ObjPtr<Fancy::Sock::HttpServerTransport> & httpServerTransport,
			int responseCode,
			int contentLength) const;

	/**
	 * Writes primitive text or html content, caring for headers like Content-Type and Content-Length.
	 * contentType defaults to FANCYSOCK_HTTPD_CONTENTTYPE_TEXT_UTF8.
	 */
	void writeResponse(
			const Fancy::ObjPtr<Fancy::Sock::HttpServerTransport> & httpServerTransport,
			int responseCode,
			const std::string & responseContent,
			const std::string & contentType="",
			const std::string & cookieToSet="",
			int responseTTL=3600) const;

        /**
         * Writes a redirect header and html note, caring for headers like Location, Content-Type and Content-Length
         */
	void writeRedirect(
			const Fancy::ObjPtr<Fancy::Sock::HttpServerTransport> & httpServerTransport,
			const std::string & redirectToUrl,
			bool permanent=true,
			int responseTTL=3600) const;

	/**
	 * Sets the base URL of the HTTP server, for use in absolute Links presented to the user.
	 * Preferrably without a trailing '/', i.e. "http://myapp.mycompany.com", not "http://myapp.mycompany.com/"
	 */
	void setBaseURL(const std::string & baseUrl);

	/**
	 * Returns whatever base URL of the application has been set via setBaseURL().
	 */
	std::string getBaseURL() const;

	/**
	 * Parses the query string (either from HttpServerTransport::getRequestQuery() or from the body of a POST request).
	 * Optionally (default yes) decodes common URL encodings like %20 for space.
	 * Values of repeated parameters are concatenated with ", " inbetween.
	 */
	Fancy::ObjPtr<Fancy::Map> parseQueryParameters(
	    const std::string & queryString,
	    bool decode=true) const;

	/**
	 * Collects up to maxBytes of the content body into localChannel.
	 *
	 * If more date are available, they are swallowed, i.e. read but thrown away, and code 413 is sent to the client.
	 *
	 * Returns the number of bytes read. May be > maxBytes, in which case the request has been handled and should not be handled any further.
	 *
	 * LocalChannel could by a BufferChannel, Base64Channel, FileChannel, ...
	 */
	int collectContentBody(
	    const Fancy::ObjPtr<Fancy::Sock::HttpServerTransport> & httpServerTransport,
	    const Fancy::ObjPtr<Fancy::Channel> & localChannel,
	    int maxBytes) const;


	/**
	 * Swallows the (rest of) the content body.
	 * Useful if an error or local problem has been detected early, to be able to swallow anything and send an error message.
	 *
	 * Returns the number of bytes swallowed.
	 */
	int swallowRest(
	    const Fancy::ObjPtr<Fancy::Sock::HttpServerTransport> & httpServerTransport) const;

	/**
	 * Tries to decode the basicAuth parts.
	 *
	 * If the request headers contain an Authorization: header for method Basic,
	 * the returned list contains the specified user and password as Fancy::Strings, already decoded from base 64.
	 *
	 * Otherwise, an empty list is returned.
	 */
	Fancy::ObjPtr<Fancy::List> decodeBasicAuthUserPassword(
	    const Fancy::ObjPtr<Fancy::Map> & requestHeaders) const;

	/**
	 * Sends a code 401 and an apropriate response header like
	 *   WWW-Authenticate: myMethod realm="myRealm"
	 * resp.
   *   WWW-Authenticate: Basic realm="fooRealm"
	 */
	void writeAuthorizationDemand(
	    const Fancy::ObjPtr<Fancy::Sock::HttpServerTransport> & httpServerTransport,
	    const std::string & method,
	    const std::string & realm,
	    int responseTTL=3600) const;

protected:

	std::string baseUrl;

};


class FANCYSOCK_API ResourceProvider : virtual public HttpDRequestHandler
{
protected:
	//ResourceProvider();
	virtual ~ResourceProvider();
public:

	typedef Fancy::ObjPtr<ResourceProvider> ptr_t;

	/*
	 * mountPath = mount point in the URL.
	 * resourcePath = local directory containing the resoures.
	 * i.e.: mountPath="/docs", resourcePath="/usr/share/myapp/docs"
	 *
	 * In the usual case of mapping a directory, do NOT append a directory seperator at the end of mountPath nor resourcePath.
	 */
	static ptr_t newInstance(
			const std::string & mountPath,
			const std::string & resourcePath);

	virtual void handleRequest(
			const Fancy::ObjPtr<Fancy::Map> & requestHeaders,
			const Fancy::ObjPtr<Fancy::Sock::HttpServerTransport> & httpServerTransport) = 0;

	virtual void setResponseTTL(int r=3600) = 0;

};


/**
 * An UploadParser processes a multipart/form-data POST.
 * Normal parameters are fed into a name/value map available through getQueryParams().
 * File names are sanitized, files written into uploadPath/
 * File infos are fed a filename/DirEntry map available through getUploadedFilesInfos().
 */
class FANCYSOCK_API UploadParser : virtual public Fancy::Object
{
protected:
    //UploadParser();
    virtual ~UploadParser();
public:

    typedef Fancy::ObjPtr<UploadParser> ptr_t;

    /**
     * Creates a new UploadParser instance
     */
    static ptr_t newInstance(
        const Fancy::ObjPtr<Fancy::Map> & requestHeaders,
        const Fancy::ObjPtr<Fancy::Sock::HttpServerTransport> & httpServerTransport,
        const std::string & uploadPath,
        bool allowOverwriting = false,
        uint8_t maxNoFiles = 1,
        uint64_t maxFileSize = 2*1024*1024);

    /**
     * Actually handles the POST data.
     */
    virtual void handleData() = 0;

    /**
     * Returns the query parameters parse from the POST data.
     */
    virtual Fancy::ObjPtr<Fancy::Map> getQueryParams() const = 0;

    /**
     * Returns the DirEntry objects for files attempted to retrieve.
     * If an error occured during the upload of a file, the file still may not exist or be empty.
     */
    virtual Fancy::ObjPtr<Fancy::Map> getUploadedFilesInfos() const = 0;

    /**
     * Returns the "recommended" return code.
     * 200 if all went ok.
     * An apropriate 4xx or even 5xx in case of typical errors.
     */
    virtual int getRecommendedResponseCode() const = 0;

    /**
     * Returns the "recommended" response text.
     * Empty or some success message if all went ok.
     * An error description otherwise.
     */
    virtual std::string getRecommendedResponseText() const = 0;
};


}}} // end namespaces

#endif /* FANCYSOCK_HTTPD_REQUEST_HANDLER_H_ */
