/***********************************************************
 * $Id$
 *
 * FancySock HTTPD helpers
 *
 * Copyright (C) 2013 ITEG IT-Engineers GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ************************************************************/

#ifndef FANCYSOCK_HTTPD_SESSION_H_
#define FANCYSOCK_HTTPD_SESSION_H_

#include <fancysock.h>

#include <fancyobject.h>
#include <fancythread.h>

//#include <map>
#include <string>
//#include <vector>


#define FANCYSOCK_HTTPD_SESSION_DEFAULT_COOKIE_NAME "SESSIONID"


namespace Fancy {

  class Map;
  class String;

  namespace Sock {
    class HttpServerTransport;

    namespace HttpD {

      /**
       * HttpD::Session provides a basic Session POJO, and helper methods to manage session Cookies and the like.
       *
       * Multithreaded applications need to use
       *   Fancy::MutexLocker sessionLocker(session->getStatusMutex());
       * before calling touch() or any setFoo() method!
       */
      class FANCYSOCK_API Session : public virtual Fancy::Object {

      protected:
        //Session();
        virtual ~Session();

      public:

        typedef Fancy::ObjPtr<Session> ptr_t;

        /**
         * Creates a new HttpSession object.
         */
        static ptr_t newInstance(
            const std::string & sessionId);

        /**
         * Returns a Fancy::Mutex, so RequestHandlers changing the mutex can easily use
         *   Fancy::MutexLocker sessionLocker(session->getStatusMutex());
         * before calling any setFoo() method.
         */
        virtual Fancy::Mutex* getStatusMutex() = 0;

        /**
         * Returns the sessionId
         */
        virtual std::string getSessionId() const = 0;

        /**
         * Returns the value for a Set-Cookie response header.
         * The expiry date is set based on sessionTimeout.
         *
         * Example:
         * SESSIONID=1234; Domain=docs.foo.com; Path=/accounts; Expires=Wed, 13 Jan 2021 22:23:01 GMT; Secure; HttpOnly
         */
        virtual std::string getCookieToSet(
            bool secure = true,
            bool httpOnly = true) const = 0;

        /**
         * Sets the login (like the userId or domain\\userId), and an optional remoteHost
         */
        virtual void setLogin(const std::string & login,
            const std::string & remoteHost="") = 0;

        /**
         * Clears the login, i.e. sets login and remoteHost it to ""
         */
        virtual void clearLogin() = 0;

        /**
         * Returns the last login set by setLogin() or clearLogin()
         */
        virtual std::string getLogin() const = 0;

        /**
         * Returns the last remoteHost set by setLogin() or clearLogin()
         */
        virtual std::string getRemoteHost() const = 0;

        /**
         * Touches the session as being used, i.e. sets lastTouchMillis to the current system time.
         */
        virtual void touch() = 0;

        /**
         * Returns lastTouchMillis
         */
        virtual uint64_t getLastTouchMillis() const = 0;

        /**
         * If lastTouchMillis is set to a past time, the milliseconds past since then is returned.
         * Otherwise, the current system time millis are returned to look like an ancient session.
         */
        virtual uint64_t getIdleMillis() const = 0;

        /**
         * Returns millis since session creation
         */
        virtual uint64_t getAgeMillis() const = 0;

        /**
         * Returns if the current system time is after lastTouchMillis + 1000*sessionTimoutSeconds.
         */
        virtual bool isTimedOut() const = 0;

        /**
         * Sets sessionTimeoutSeconds
         */
        virtual void setSessionTimeoutSeconds(uint32_t sessionTimeoutSeconds) = 0;

        /**
         * Returns the last sessionTimeoutSeconds set by setSessionTimeoutSeconds()
         */
        virtual uint32_t getSessionTimeoutSeconds() const = 0;

        /**
         * Sets cookieName to use.
         * If sessionCookieName is specified empty here, it set to FANCYSOCK_HTTPD_SESSION_DEFAULT_COOKIE_NAME = "SESSIONID" instead.
         */
        virtual void setCookieName(const std::string & cookieName) = 0;

        /**
         * Returns the cookieName, be it the default or set by setCookieName()
         */
        virtual std::string getCookieName() const = 0;

        /**
         * Sets cookieDomain for usage by addSetCookie()
         */
        virtual void setCookieDomain(const std::string & cookieDomain) = 0;

        /**
         * Returns cookieDomain set by last call to setCookieDomain()
         */
        virtual std::string getCookieDomain() const = 0;

        /**
         * Sets cookiePath for usage by addSetCookie()
         */
        virtual void setCookiePath(const std::string & cookiePath) = 0;

        /**
         * Returns cookiePath set by last call to setCookieDomain()
         */
        virtual std::string getCookiePath() const = 0;

        /**
         * Returns an ObjPtr to the non-null Fancy::Map sessionProperties, to allow storing further properties
         */
        virtual Fancy::ObjPtr<Fancy::Map> getSessionProperties() const = 0;

      };


      /**
       * HttpD::Session provides a basic Session POJO, and helper methods to manage session Cookies and the like.
       */
      class FANCYSOCK_API SessionManager : public virtual Fancy::Object {

      protected:
        //SessionManager();
        virtual ~SessionManager();

      public:

        typedef Fancy::ObjPtr<SessionManager> ptr_t;

        /**
         * Creates a new HttpSession object.
         *
         * If defaultCookieName is specified empty, the default is used instead,
         * FANCYSOCK_HTTPD_SESSION_DEFAULT_COOKIE_NAME = "SESSIONID".
         */
        static ptr_t newInstance(
            uint32_t            defaultSessionTimeoutSeconds = 30*60,
            const std::string & defaultCookieName = "",
            const std::string & defaultCookieDomain = "",
            const std::string & defaultCookiePath = "");

        /**
    	 * @return An optional URL, which is set, if calls to this serives neead anauthenticated
    	 *         user in the HTTP session.
    	 */
    	virtual ObjPtr<String> getLoginUrl() const = 0;


    	/**
    	 * Set the optional login URL, which is non-null, if only authenticated users
    	 * are allowed to call this service.
    	 */
    	virtual void setLoginUrl(const ObjPtr<String> &loginUrl) = 0;

        /**
         * Returns an eventual known session from the internal session list,
         * or null if the client did not send a sessionId or the session could not be found.
         *
         * For environments that don't support cookies, use something else as sessionId and retrieveSession(string sessionId) below.
         */
        virtual Fancy::ObjPtr<Session> retrieveSession(
            const Fancy::ObjPtr<Fancy::Map> & requestHeaders) = 0;

        /**
         * Returns an eventual known session from the internal session list,
         * or a new session if the client did not send a sessionId, or the session could not be found, or the session has timed out.
         *
         * For environments that don't support cookies, use something else as sessionId and retrieveOrCreateSession(string sessionId) below.
         */
        virtual Fancy::ObjPtr<Session> retrieveOrCreateSession(
            const Fancy::ObjPtr<Fancy::Map> & requestHeaders) = 0;

        /**
         * Tries to retrieve the sessionId from a cookie named cookieName and returns it.
         *
         * If the request headers do not contain an according cookie, an empty string is returned.
         */
        virtual std::string getSessionIdFromCookie(
            const Fancy::ObjPtr<Fancy::Map> & requestHeaders,
            const std::string & cookieName = "") const = 0;

        /**
         * Returns an eventual known session from the internal session list,
         * or a new session if the session could not be found, or the session has timed out.
         *
         * This should be used only for environments that don't support cookies, otherwise use retrieveOrCreateSession(Map headers).
         */
        virtual Fancy::ObjPtr<Session> retrieveOrCreateSession(
      		  const std::string & sessionId) = 0;

        /**
         * Returns an eventual known session from the internal session list,
         * or null if the session could not be found.
         *
         * This should be used only for environments that don't support cookies, otherwise use retrieveSession(Map headers).
         */
        virtual Fancy::ObjPtr<Session> retrieveSession(
      		  const std::string & sessionId) = 0;

      };


    }}} // end namespaces

#endif /* FANCYSOCK_HTTPD_SESSION_H_ */
