/***********************************************************
 * $Id$
 *
 * FancyLibs JSON-RPC-2.0 support
 *
 * Copyright (C) 2014 ITEG IT-Engineers GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ************************************************************/

#ifndef FANCSOCKJSONRPC2HANDLER_H_
#define FANCSOCKJSONRPC2HANDLER_H_

#include <fancysockexception.h>
#include <fancyserializable.h>
#include <fancyreflectiontarget.h>

namespace Fancy {

class Channel;
class JsonOptions;
class SerializableFactory;
class TypeConverter;
class Number;
class String;
template<class T> class ObjectArray;

namespace Sock {

class JsonRpc2Error;
/**
 * A rpc call is represented by sending a Request object to a Server.
 * The Request object has the following members:
 * -jsonrpc
 *    A String specifying the version of the JSON-RPC protocol. MUST be exactly "2.0".
 * - method
 *    A String containing the name of the method to be invoked. Method names that begin
 *    with the word rpc followed by a period character (U+002E or ASCII 46) are reserved
 *    for rpc-internal methods and extensions and MUST NOT be used for anything else.
 * - params
 *    A Structured value that holds the parameter values to be used during the invocation
 *    of the method. This member MAY be omitted.
 * - id
 *    An identifier established by the Client that MUST contain a String, Number, or NULL
 *    value if included. If it is not included it is assumed to be a notification.
 *    The value SHOULD normally not be Null [1] and Numbers SHOULD NOT contain
 *    fractional parts [2].
 *
 * The Server MUST reply with the same value in the Response object if included.
 * This member is used to correlate the context between the two objects.
 */
class JsonRpc2Request : virtual public Fancy::Serializable {

	Fancy::ObjPtr<Fancy::String> jsonrpc;
	Fancy::ObjPtr<Fancy::String> method;
	Fancy::ObjPtr<Fancy::ObjectArray<Fancy::Object> > params;
	Fancy::ObjPtr<Fancy::Object> id;

public:
	typedef ObjPtr<JsonRpc2Request> ptr_t;

	static const char *getSerialID_s();
	static int64_t getSerialVersionUID_s();

	const char *getSerialID() const;
	int64_t getSerialVersionUID() const;

	static ptr_t newInstance();
	static Serializable::ptr_t newSerializableInstance();

	const Fancy::ObjPtr<Fancy::Object>& getId() const;
	void setId(const Fancy::ObjPtr<Fancy::Object>& id);
	const Fancy::ObjPtr<Fancy::String>& getJsonrpc() const;
	void setJsonrpc(const Fancy::ObjPtr<Fancy::String>& jsonrpc);
	const Fancy::ObjPtr<Fancy::String>& getMethod() const;
	void setMethod(const Fancy::ObjPtr<Fancy::String>& method);
	const Fancy::ObjPtr<Fancy::ObjectArray<Fancy::Object> >& getParams() const;
	void setParams(const Fancy::ObjPtr<Fancy::ObjectArray<Fancy::Object> >& params);

	virtual void writeFields(const ObjPtr<Serializer> &serializer) const;
    virtual ObjPtr<Serializable> readFields(const ObjPtr<Serializer> &serializer, int64_t serialVersionUID);

    virtual std::string toString() const;

protected:
	JsonRpc2Request();
	~JsonRpc2Request();
};

/**
 * When a rpc call is made, the Server MUST reply with a Response, except for in the case of Notifications. The Response is expressed as a single JSON Object, with the following members:
 *
 * - jsonrpc
 *    A String specifying the version of the JSON-RPC protocol. MUST be exactly "2.0".
 * - result
 *    This member is REQUIRED on success.
 *    This member MUST NOT exist if there was an error invoking the method.
 *    The value of this member is determined by the method invoked on the Server.
 * - error
 *    This member is REQUIRED on error.
 *    This member MUST NOT exist if there was no error triggered during invocation.
 *    The value for this member MUST be an Object as defined in section 5.1.
 * - id
 *    This member is REQUIRED.
 *    It MUST be the same as the value of the id member in the Request Object.
 *    If there was an error in detecting the id in the Request object (e.g. Parse error/Invalid Request), it MUST be Null.
 *    Either the result member or error member MUST be included, but both members MUST NOT be included.
 */
class JsonRpc2Response : virtual public Fancy::Serializable {

	Fancy::ObjPtr<Fancy::String> jsonrpc;
	Fancy::ObjPtr<Fancy::Object> result;
	Fancy::ObjPtr<JsonRpc2Error> error;
	Fancy::ObjPtr<Fancy::Object> id;

public:
	typedef ObjPtr<JsonRpc2Response> ptr_t;

	static const char *getSerialID_s();
	static int64_t getSerialVersionUID_s();

	const char *getSerialID() const;
	int64_t getSerialVersionUID() const;

	static ptr_t newInstance();
	static Serializable::ptr_t newSerializableInstance();

	const Fancy::ObjPtr<JsonRpc2Error>& getError() const;
	void setError(const Fancy::ObjPtr<JsonRpc2Error>& error);
	const Fancy::ObjPtr<Fancy::Object>& getId() const;
	void setId(const Fancy::ObjPtr<Fancy::Object>& id);
	const Fancy::ObjPtr<Fancy::String>& getJsonrpc() const;
	void setJsonrpc(const Fancy::ObjPtr<Fancy::String>& jsonrpc);
	const Fancy::ObjPtr<Fancy::Object>& getResult() const;
	void setResult(const Fancy::ObjPtr<Fancy::Object>& result);

	virtual void writeFields(const ObjPtr<Serializer> &serializer) const;
    virtual ObjPtr<Serializable> readFields(const ObjPtr<Serializer> &serializer, int64_t serialVersionUID);

    virtual std::string toString() const;

protected:
    JsonRpc2Response();
    ~JsonRpc2Response();
};

/**
 * When a rpc call encounters an error, the Response Object MUST contain the error member with a value that is a Object with the following members:
 *
 * - code
 *    A Number that indicates the error type that occurred.
 *    This MUST be an integer.
 * - message
 *    A String providing a short description of the error.
 *    The message SHOULD be limited to a concise single sentence.
 * - data
 *    A Primitive or Structured value that contains additional information about the error.
 *    This may be omitted.
 *    The value of this member is defined by the Server
 *    (e.g. detailed error information, nested errors etc.).
 *    The error codes from and including -32768 to -32000 are reserved
 *    for pre-defined errors. Any code within this range, but not defined
 *    explicitly below is reserved for future use. The error codes are nearly
 *    the same as those suggested for XML-RPC at the following
 *    url: http://xmlrpc-epi.sourceforge.net/specs/rfc.fault_codes.php
 *
 * The following error codes are predefined by the JSON-RPC Specification:
 *    - 32700	Parse error	Invalid JSON was received by the server.
 *              An error occurred on the server while parsing the JSON text.
 *    - 32600	Invalid Request	The JSON sent is not a valid Request object.
 *    - 32601	Method not found	The method does not exist / is not available.
 *    - 32602	Invalid params	Invalid method parameter(s).
 *    - 32603	Internal error	Internal JSON-RPC error.
 *    - 32000 to -32099	Server error	Reserved for implementation-defined server-errors.
 *    The remainder of the space is available for application defined errors.
 */
class JsonRpc2Error : virtual public Fancy::Serializable {

	Fancy::ObjPtr<Fancy::Number> code;
	Fancy::ObjPtr<Fancy::String> message;
	Fancy::ObjPtr<Fancy::Object> data;

public:

	/**
	 * <code>Parse error</code>:
	 * Invalid JSON was received by the server. An error occurred on the server while parsing the JSON text.
	 */
	static const int JSON_RPC_PARSE_ERROR = -32700;

	/**
	 * <code>Invalid Request</code>:
	 * The JSON sent is not a valid Request object.
	 */
	static const int JSON_RPC_INVALID_REQUEST = -32600;

	/**
	 * <code>Method not found</code>:
	 * The method does not exist / is not available.
	 */
	static const int JSON_RPC_METHOD_NOT_FOUND = -32601;

	/**
	 * <code>Invalid params</code>:
	 * Invalid method parameter(s).
	 */
	static const int JSON_RPC_INVALID_PARAMS = -32602;

	/**
	 * <code>Internal error</code>:
	 * Internal JSON-RPC error.
	 */
	static const int JSON_RPC_INTERNAL_ERROR = -32603;

	/**
	 * <code>Server error</code>:
	 * Reserved for implementation-defined server-errors.
	 */
	static const int JSON_RPC_SERVER_ERROR_0 = -32000;

	/**
	 * <code>Server error</code>:
	 * Reserved for implementation-defined server-errors.
	 */
	static const int JSON_RPC_SERVER_ERROR_1 = -32001;

	/**
	 * <code>Server error</code>:
	 * Reserved for implementation-defined server-errors.
	 */
	static const int JSON_RPC_SERVER_ERROR_2 = -32002;

	/**
	 * <code>Server error</code>:
	 * Reserved for implementation-defined server-errors.
	 */
	static const int JSON_RPC_SERVER_ERROR_3 = -32003;

	/**
	 * <code>Server error</code>:
	 * Reserved for implementation-defined server-errors.
	 */
	static const int JSON_RPC_SERVER_ERROR_4 = -32004;

	/**
	 * <code>Server error</code>:
	 * Reserved for implementation-defined server-errors.
	 */
	static const int JSON_RPC_SERVER_ERROR_5 = -32005;

	/**
	 * <code>Server error</code>:
	 * Reserved for implementation-defined server-errors.
	 */
	static const int JSON_RPC_SERVER_ERROR_6 = -32006;

	/**
	 * <code>Server error</code>:
	 * Reserved for implementation-defined server-errors.
	 */
	static const int JSON_RPC_SERVER_ERROR_7 = -32007;

	/**
	 * <code>Server error</code>:
	 * Reserved for implementation-defined server-errors.
	 */
	static const int JSON_RPC_SERVER_ERROR_8 = -32008;

	/**
	 * <code>Server error</code>:
	 * Reserved for implementation-defined server-errors.
	 */
	static const int JSON_RPC_SERVER_ERROR_9 = -32009;

	/**
	 * <code>Server error</code>:
	 * Reserved for implementation-defined server-errors.
	 */
	static const int JSON_RPC_SERVER_ERROR_99 = -32099;


	typedef ObjPtr<JsonRpc2Error> ptr_t;

	static const char *getSerialID_s();
	static int64_t getSerialVersionUID_s();

	const char *getSerialID() const;
	int64_t getSerialVersionUID() const;

	static ptr_t newInstance();
	static Serializable::ptr_t newSerializableInstance();

	const Fancy::ObjPtr<Fancy::Number>& getCode() const;
	void setCode(const Fancy::ObjPtr<Fancy::Number>& code);
	const Fancy::ObjPtr<Fancy::Object>& getData() const;
	void setData(const Fancy::ObjPtr<Fancy::Object>& data);
	const Fancy::ObjPtr<Fancy::String>& getMessage() const;
	void setMessage(const Fancy::ObjPtr<Fancy::String>& message);

	virtual void writeFields(const ObjPtr<Serializer> &serializer) const;
    virtual ObjPtr<Serializable> readFields(const ObjPtr<Serializer> &serializer, int64_t serialVersionUID);

    virtual std::string toString() const;

protected:
    JsonRpc2Error();
    ~JsonRpc2Error();
};



/**
 * Exception thrown, when an error is received by the client from the server.
 */
class FANCYSOCK_API JsonRpc2Exception : public Fancy::Sock::Exception
{
	JsonRpc2Error::ptr_t jsonRpcError;
public:
	JsonRpc2Exception (fancylog_service *service,
                 int facility,
                 int priority,
                 const JsonRpc2Error::ptr_t &jpe,
                 const char *fmt, ...) throw()
#ifdef __GNUG__
    __attribute__((format(printf,6,7)))
#endif
    ;

	JsonRpc2Exception (int facility,
                 int priority,
                 const JsonRpc2Error::ptr_t &jpe,
                 const char *fmt, ...) throw()
#ifdef __GNUG__
    __attribute__((format(printf,5,6)))
#endif
    ;

  virtual ~JsonRpc2Exception() throw();

  JsonRpc2Error::ptr_t getJsonRpcError() const throw();
};


class JsonRpc2Handler : virtual public Fancy::ReflectionTarget {

protected:
	JsonRpc2Handler();
	virtual ~JsonRpc2Handler();

public:
	typedef ObjPtr<JsonRpc2Handler> ptr_t;

	/**
	 * Instantiate a new JSON RPC handler using a given reflection target.
	 */
	static ptr_t newInstance(
			const ObjPtr<Channel> &channel,
			const ObjPtr<SerializableFactory> &sf,
			const ObjPtr<JsonOptions> &options,
			const ObjPtr<ReflectionTarget> &target);

	/**
	 * Instantiate a new JSON RPC handler using a given reflection target.
	 * This variant allows to register a customized type converter for
	 * deserializing stringified enum instances.
	 */
	static ptr_t newInstance(
			const ObjPtr<Channel> &channel,
			const ObjPtr<SerializableFactory> &sf,
			const ObjPtr<TypeConverter> &tc,
			const ObjPtr<JsonOptions> &options,
			const ObjPtr<ReflectionTarget> &target);


	/**
	 * Instantiate a new server instance, which may handle JSON-RPC calls
	 * inside @c handleInvocation()
	 */
	template<class Interface>
	static ptr_t newInstanceT(const ObjPtr<Channel> &channel,
			const ObjPtr<SerializableFactory> &sf,
			const ObjPtr<JsonOptions> &options,
			const ObjPtr<Interface> &target) {
		return newInstance(channel,sf,options,Interface::reflect(target));
	}

	/**
	 * Instantiate a new server instance, which may handle JSON-RPC calls
	 * inside @c handleInvocation()
	 * This variant allows to register a customized type converter for
	 * deserializing stringified enum instances.
	 */
	template<class Interface>
	static ptr_t newInstanceT(const ObjPtr<Channel> &channel,
			const ObjPtr<SerializableFactory> &sf,
			const ObjPtr<TypeConverter> &tc,
			const ObjPtr<JsonOptions> &options,
			const ObjPtr<Interface> &target) {
		return newInstance(channel,sf,tc,options,Interface::reflect(target));
	}

	/**
	 * Instantiate a new client instance, which issues JSON-RPC calls
	 * to the given channel.
	 */
	template<class Interface>
	static ptr_t newInstanceT(const ObjPtr<Channel> &channel,
			const ObjPtr<SerializableFactory> &sf,
			const ObjPtr<JsonOptions> &options) {
		return newInstance(channel,sf,options,Interface::reflect(ObjPtr<Interface>()));
	}

	/**
	 * Instantiate a new client instance, which issues JSON-RPC calls
	 * to the given channel.
	 * This variant allows to register a customized type converter for
	 * deserializing stringified enum instances.
	 */
	template<class Interface>
	static ptr_t newInstanceT(const ObjPtr<Channel> &channel,
			const ObjPtr<SerializableFactory> &sf,
			const ObjPtr<TypeConverter> &tc,
			const ObjPtr<JsonOptions> &options) {
		return newInstance(channel,sf,tc,options,Interface::reflect(ObjPtr<Interface>()));
	}

	/**
	 * Instantiate a new client instance, which issues JSON-RPC calls
	 * to the given channel.
	 */
	template<class Interface>
	static typename Interface::ptr_t newClientProxy(const ObjPtr<Channel> &channel,
			const ObjPtr<SerializableFactory> &sf,
			const ObjPtr<JsonOptions> &options) {
		return Interface::newProxyInstance(newInstance(channel,sf,options,Interface::reflect(ObjPtr<Interface>())));
	}

	/**
	 * Instantiate a new client instance, which issues JSON-RPC calls
	 * to the given channel.
	 */
	template<class Interface>
	static typename Interface::ptr_t newClientProxy(const ObjPtr<Channel> &channel,
			const ObjPtr<SerializableFactory> &sf,
			const ObjPtr<TypeConverter> &tc,
			const ObjPtr<JsonOptions> &options) {
		return Interface::newProxyInstance(newInstance(channel,sf,tc,options,Interface::reflect(ObjPtr<Interface>())));
	}

	/**
	 * Server-side interface, part 0.
	 * Read an incoming invocation.
	 */
	virtual JsonRpc2Request::ptr_t readRequest() = 0;

	/**
	 * Server-side interface, part 1.
	 * Invoke target function and return the response  object.
	 */
	virtual JsonRpc2Response::ptr_t handleRequest(const JsonRpc2Request::ptr_t &request) = 0;

	/**
	 * Server-side interface, part 0 & 1 combined.
	 * Wait for an incoming invocation, invoke target function and
	 * return the response  object.
	 */
	virtual JsonRpc2Response::ptr_t handleInvocation() = 0;

	/**
	 * Server-side interface, part 2.
	 * Write a response object to the underlying channel.
	 */
	virtual void writeResponse(const JsonRpc2Response::ptr_t &response) = 0;
};

} /* namespace Sock */
} /* namespace Fancy */

#endif /* FANCSOCKJSONRPC2HANDLER_H_ */
