/***********************************************************
 * $Id$
 *
 * FancyLibs
 *
 * Copyright (C) 2016 ITEG IT-Engineers GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ************************************************************/

#ifndef FANCYSOCKNTPSTATUS_H_
#define FANCYSOCKNTPSTATUS_H_

#include <fancysock.h>
#include <fancyobject.h>

namespace Fancy {

class ClockVal;

namespace Sock {

/**
 * A simple NTP status generated from an NTP query.
 * The detailed format can be found under: https://tools.ietf.org/html/rfc5905
 *
 *  |LI | VN  |Mode |    Stratum    |     Poll      |   Precision    |
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *  |                          Root  Delay                           |
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *  |                       Root  Dispersion                         |
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *  |                     Reference Identifier                       |
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *  |                                                                |
 *  |                    Reference Timestamp (64)                    |
 *  |                                                                |
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *  |                                                                |
 *  |                    Originate Timestamp (64)                    |
 *  |                                                                |
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *  |                                                                |
 *  |                     Receive Timestamp (64)                     |
 *  |                                                                |
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *  |                                                                |
 *  |                     Transmit Timestamp (64)                    |
 *  |                                                                |
 *  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 */
class FANCYSOCK_API NtpStatus: public virtual Fancy::Object {

    int li;
    int vn;
    int mode;
    int stratum;
    int poll;
    int precision;
    double rootDelay;
    double rootDispersion;
    uint32_t referenceIdentifier;
    uint64_t referenceTimestamp;
    uint64_t originTimestamp;
    uint64_t receiveTimestamp;
    uint64_t transmitTimestamp;

    uint64_t finalTimestamp;
    double requestDuration;

protected:
    NtpStatus();
	virtual ~NtpStatus();

public:

	typedef Fancy::ObjPtr<NtpStatus> ptr_t;

    static const int NTP_MODE_Symmetric_Active = 1;
    static const int NTP_MODE_Symmetric_Passive = 2;
    static const int NTP_MODE_Client = 3;
    static const int NTP_MODE_Server = 4;
    static const int NTP_MODE_Broadcast_Server = 5;
    static const int NTP_MODE_Broadcast_Client = 6;

    static const int NTP_LI_OK = 0; // no warning
    static const int NTP_LI_MINUTE_61 = 1; // last minute of the day has 61 seconds
    static const int NTP_LI_MINUTE_59 = 2; // last minute of the day has 59 seconds
    static const int NTP_LI_UNSYNCHRONIZED = 3; // unknown (clock unsynchronized)

    /**
     * Generate a new empty instance with mode set to @c NTP_MODE_Client.
     */
    static ptr_t newInstance();

    /**
     * Parse an 48 byte NTP response, which has been initiated at the given
     * clock instance. This call initializes @c finalTimestamp from the current date/time
     * and the request duration from the clock difference between start and the
     * current clock value.
     */
    static ptr_t newInstance(const Fancy::ClockVal &start, const void *response);

    /**
     * Serialize to a 48 byte buffer.
     */
    void serialize(void *buf);

    /**
     * Return the calculated round trip delay in seconds.
     */
    double getRoundtripDelay() const;

    /**
     * Return the calculated time offset between the client and the server in seconds.
     * Positive means client is ahead of server time.
     */
    double getTimeOffset() const;

    /**
     * Return the calculated unsynchronized period of the server aka
     * @c receiveTimestamp - referenceTimestamp
     */
    double getUnsynchronizedPeriod() const;

    int getLi() const {
        return this->li;
    }

    void setLi(int li) {
        this->li = li;
    }

    int getMode() const {
        return this->mode;
    }

    void setMode(int mode) {
        this->mode = mode;
    }

    uint64_t getOriginTimestamp() const {
        return this->originTimestamp;
    }

    void setOriginTimestamp(uint64_t originTimestamp) {
        this->originTimestamp = originTimestamp;
    }

    int getPoll() const {
        return this->poll;
    }

    void setPoll(int poll) {
        this->poll = poll;
    }

    int getPrecision() const {
        return this->precision;
    }

    void setPrecision(int precision) {
        this->precision = precision;
    }

    uint64_t getReceiveTimestamp() const {
        return this->receiveTimestamp;
    }

    void setReceiveTimestamp(uint64_t receiveTimestamp) {
        this->receiveTimestamp = receiveTimestamp;
    }

    uint32_t getReferenceIdentifier() const {
        return this->referenceIdentifier;
    }

    void setReferenceIdentifier(uint32_t referenceIdentifier) {
        this->referenceIdentifier = referenceIdentifier;
    }

    uint64_t getReferenceTimestamp() const {
        return this->referenceTimestamp;
    }

    void setReferenceTimestamp(uint64_t referenceTimestamp) {
        this->referenceTimestamp = referenceTimestamp;
    }

    double getRootDelay() const {
        return this->rootDelay;
    }

    void setRootDelay(double rootDelay) {
        this->rootDelay = rootDelay;
    }

    double getRootDispersion() const {
        return this->rootDispersion;
    }

    void setRootDispersion(double rootDispersion) {
        this->rootDispersion = rootDispersion;
    }

    int getStratum() const {
        return this->stratum;
    }

    void setStratum(int stratum) {
        this->stratum = stratum;
    }

    uint64_t getTransmitTimestamp() const {
        return this->transmitTimestamp;
    }

    void setTransmitTimestamp(uint64_t transmitTimestamp) {
        this->transmitTimestamp = transmitTimestamp;
    }

    int getVn() const {
        return this->vn;
    }

    void setVn(int vn) {
        this->vn = vn;
    }

    uint64_t getFinalTimestamp() const {
        return this->finalTimestamp;
    }

    void setFinalTimestamp(uint64_t finalTimestamp) {
        this->finalTimestamp = finalTimestamp;
    }

    double getRequestDuration() const {
        return this->requestDuration;
    }

    void setRequestDuration(double requestDuration) {
        this->requestDuration = requestDuration;
    }
};

}

}

#endif /* FANCYSOCKNTPSTATUS_H_ */
