/***********************************************************
 * $Id$
 *
 * FancyLibs
 *
 * Copyright (C) 2009-2011 ev-i GmbH & IBCL e.U.
 * Copyright (C) 2011-2013 ITEG IT-Engineers GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ************************************************************/

/****************************************************************************
 **
 ** Author: Christoph Lechleitner / ITEG
 **
 ** Interface for wrappers for WLAN steering
 **
 *****************************************************************************/

#ifndef FANCYSOCK_WLAN_H
#define FANCYSOCK_WLAN_H

#define FANCYSOCK_WLAN_SIGNALSTRENGTH_OFFLINE -65535

#include <fancysock.h>

#include <fancythread.h>

//#include <string>

namespace Fancy {
class List;
class Map;

namespace Sock {

class Address;


/**
 * Represents a WLAN status
 */
class FANCYSOCK_API WLANStatus : virtual public Fancy::Object
{
public:
    typedef Fancy::ObjPtr<WLANStatus> ptr_t;

    virtual bool isWLANOn() = 0;

    virtual bool isWLANConnected() = 0;

    virtual int getSignalStrength() = 0;

};



/**
 * A WLAN status callback receives status updates from a WLANController (see below).
 */
class FANCYSOCK_API WLANStatusCallback : virtual public Fancy::Object
{
public:
    typedef Fancy::ObjPtr<WLANStatusCallback> ptr_t;

    /**
     * Creates a simple instance just logging callbacks.
     */
    static ptr_t newSimpleInstance();

    /**
     * Called after a status change.
     */
    virtual void onWLANStatusChange(
    		const Fancy::ObjPtr<WLANStatus> & status) = 0;

protected:

};


/**
 * A WLANController uses the OS's default technology to steer and observe WLAN connections
 * For now only Linux is supported, calling ifup/ifdown and the like.
 */
class FANCYSOCK_API WLANController : virtual public Fancy::Runnable
{
public:
    typedef Fancy::ObjPtr<WLANController> ptr_t;

#ifndef WIN32

#define FANCYWLAN_OPTION_INTERFACESFILE "INTERFACES"
// backward comp.
#define FANCY_WLAN_OPTION_INTERFACESFILE "INTERFACES"

    /**
     * Instantiates a new WLANController instance.
     *
     * interfaceName examples: wlan0
     *
     * Options supported by Linux instances:
     *
     * "INTERFACES" (FANCYSOCKWLAN_OPTION_INTERFACESFILE)
     *   Name of alternative file to use instead of /etc/network/interfaces.
     *   Example: "/etc/network/interfaces.siteXY"
     */
    static ptr_t newLinuxInstance(const std::string & interfaceName);
#endif

#define FANCYWLAN_OPTION_WLANSTATUSFILE        "WLAN-STATUS-FILE"

#define FANCYFAKEWLAN_OPTION_CONNECTDELAYMS    "CONNECT-DELAY-MILLIS"
#define FANCYFAKEWLAN_OPTION_DISCONNECTDELAYMS "DISCONNECT-DELAY-MILLIS"
//#define FANCYFAKEWLAN_OPTION_WLANUPFILE      "WLAN-UP-FILE"

    /**
     * Instantiates a faking WLANController instance.
     *
     * Options supported by Fake instances:
     *
     * "CONNECT-DELAY-MILLIS" (FANCYFAKEWLAN_OPTION_CONNECTDELAYMS):
     *   Artifical delay between connect() call and successful connection in ms.
     *   Default: 2500 ms.
     *   Negative value: Connection attempt will fail.
     *
     * "DISCONNECT-DELAY-MILLIS" (FANCYFAKEWLAN_OPTION_DISCONNECTDELAYMS):
     *   Artifical delay between disconnect() call and successful disconnection in ms.
     *   Default: 1000 ms.
     *   Negative value: No automatic disconnection.
     */
    static ptr_t newFakeInstance();

    /**
     * Must be called before (next) connect() call.
     *
     * Options supported by all instances:
     *
     * "WLAN-STATUS-FILE" (FANCYWLAN_OPTION_WLANSTATUSFILE):
     *   Name of file which's content specifies the status of the pseudo connection.
     *   If [DIS]CONNECT-DELAY-MILLIS is >= 0, it is automatically created/updated.
     *   Default:
     *     /tmp/fancysockwlan-fake-status.txt
     *   resp.
     *     %TEMP%\fancysockwlan-fake-status.txt
     * The format is shared between the fake instance and what
     * the status detection script of the Linux implementation creates:
     *   status=not connected
     * or
     *   status=connected
     *   signal level=-13
     *   link quality=42/100
     *   essid=FakeWifi
     *   access point=01:02:03:0A:0B:0C
     */
    virtual void setConnectionOptions(
    		const Fancy::ObjPtr<Fancy::Map> &options) = 0;

    /**
     * Optional. Allows to set a connectionName which is used for logging and evtl. the statusFileName.
     * This is useful to distinct log entries from multiple WLANController instances.
     *
     * Some implementations might set a connectionName that's useful enough already.
     */
    virtual void setConnectionName(
        const std::string &connName,
        bool overwriteStatusFileName=false) = 0;

    /**
     * Tries to bring up a WLAN connection.
     * Expect the actual connection process to need between 1 and 90 seconds.
     */
    virtual void connect() = 0;

    /**
     * Tries to bring down a WLAN connection.
     * Expect the actual disconnection process to need between 0 and 10 seconds.
     */
    virtual void disconnect() = 0;

    /**
     * Tells the WLAN to re-check it's connectivity ASAP.
     */
    virtual void recheckConnection() = 0;

    /**
     * Returns wether the WLAN interface is _considered_ to be up.
     */
    virtual bool isWLANSwitchedOn() = 0;

    /**
     * Returns wether the connection is _considered_ to be ok.
     * Does not guarantee real connectivity.
     */
    virtual bool isWLANConnected() = 0;

    /**
     * Must be called before starting the WLANController.
     * Sets the interval of connectivity checks (while connection considered up), and optional ping parameters.
     * Technology dependent checks are always used first, if available.
     * If this does not failed and pingAddress is set non-null, a ping is performed.
     * Setting checkIntervalSeconds <= 0 disables regular checks
     */
    virtual void setCheckOptions(
            unsigned checkIntervalSeconds = 30,
            const Fancy::ObjPtr<Address> &pingAddress = 0,
            int pingMaxHops = 15,
            int pingTimeoutMillis = 1000,
            int pingNrPings = 1,
            int pingAttempts = 3) = 0;

    /**
     * Sets a callback that receives status updates
     */
    virtual void setStatusCallback(
            const Fancy::ObjPtr<WLANStatusCallback> &statusCallback) = 0;


protected:
    WLANController();
    virtual ~WLANController();

};

}} // end namespace Fancy::Sock

#endif // ! FANCYSOCK_WLAN_H
