/***********************************************************
 * $Id$
 *
 * FancyLibs
 *
 * Copyright (C) 2009-2011 ev-i GmbH & IBCL e.U.
 * Copyright (C) 2011-2017 ITEG IT-Engineers GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ************************************************************/

#ifndef FANCYSOCKWSTRANSPORT_H_
#define FANCYSOCKWSTRANSPORT_H_

#include <fancysock.h>
#include <fancychannel.h>
#include <fancysockexception.h>

#include <string>


namespace Fancy {

class Iterator;


namespace Sock {

namespace HttpD {
class Session;
}

class Device;
class HttpServerTransport;


// Constants as defined by RFC 6455
#define FANCYSOCK_WS_VERSION_STR                 "13"
#define FANCYSOCK_WS_HANDSHAKE_GUID              "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
#define FANCYSOCK_WS_CTRLFRAME_MAX_PAYLOAD_LEN   125

// Status codes defined by RFC 6455
#define FANCYSOCK_WS_STATUS_NORMAL_CLOSE        1000  // Normal close
#define FANCYSOCK_WS_STATUS_GOING_AWAY          1001  // Other side is going away
#define FANCYSOCK_WS_STATUS_PROTOCOL_ERROR      1002  // Protocol error
#define FANCYSOCK_WS_STATUS_UNSUPPORTED_DATA    1003  // Data type not supported
#define FANCYSOCK_WS_STATUS_NONE                1005  // No status code specified in close message
#define FANCYSOCK_WS_STATUS_UNKNOWN             1006  // Unknown status since no close message was received
#define FANCYSOCK_WS_STATUS_WRONG_DATA_TYPE     1007  // Received data and specified data type mismatch
#define FANCYSOCK_WS_STATUS_POLICY_VIOLATION    1008  // Received message violates policy
#define FANCYSOCK_WS_STATUS_TOO_BIG             1009  // Received message is too big
#define FANCYSOCK_WS_STATUS_MISSING_EXT         1010  // Required extensions not supported by the server
#define FANCYSOCK_WS_STATUS_UNEXPECTED_COND     1011  // Unexpected condition
#define FANCYSOCK_WS_STATUS_TLS_HS_FAILED       1015  // Failed to perform TLS handshake

// Status codes which must not be set as status code in close frames
#define FANCYSOCK_WS_NON_CLOSE_FRAME_STATI  \
  { FANCYSOCK_WS_STATUS_NONE, FANCYSOCK_WS_STATUS_UNKNOWN, FANCYSOCK_WS_STATUS_TLS_HS_FAILED }

// Default values
#define FANCYSOCK_WS_DEF_DATA_TYPE_WR        BinaryData
#define FANCYSOCK_WS_DEF_MAX_PAYLOAD_LEN     4*1024*1024  // 4 MB
#define FANCYSOCK_WS_DEF_CLOSE_ACK_TIMEOUT   30000        // 30 sec

// Internal defines
#ifdef __GNUG__
#  define _FANCYSOCK_WS_ATTR_FMT_PRINTF(strIdx, firstToCheck) \
     __attribute__((format(printf, strIdx, firstToCheck)))
#else
#  define _FANCYSOCK_WS_ATTR_FMT_PRINTF(strIdx, firstToCheck)
#endif


typedef uint16_t wsStatusCode_t;


/**
 * WebSocket specific exception class.
 */
class FANCYSOCK_API WsException: public Exception {

protected:
  inline WsException(int facility, int priority) throw():
    Exception(log_service, facility, priority)
  {
  }

public:
  inline WsException(int facility, int priority,
                     const char *fmt, ...) throw() _FANCYSOCK_WS_ATTR_FMT_PRINTF(4, 5):
    Exception(log_service, facility, priority)
  {
    va_list ap;
    va_start(ap, fmt);
    appendMessageV(fmt, ap);
    va_end(ap);
  }

  inline WsException(int facility, int priority, const char *msg,
                     va_list vargs) throw() _FANCYSOCK_WS_ATTR_FMT_PRINTF(4, 0):
    Exception(facility, priority, msg, vargs)
  {
  }
};

/**
 * WebSocket specific exception thrown if socket has been closed unexpectedly.
 */
class FANCYSOCK_API WsCloseException: public WsException {

  wsStatusCode_t statusCode;
  std::string reason;

public:
  inline WsCloseException(int facility, int priority,
                          wsStatusCode_t statusCode = FANCYSOCK_WS_STATUS_UNKNOWN,
                          const std::string &reason = "") throw():
    WsException(facility, priority, "WebSocket closed (Status code: %u)! %s",
      static_cast<unsigned>(statusCode), reason.c_str()),
    statusCode(statusCode), reason(reason)
  {
  }

  WsCloseException(int facility, int priority, wsStatusCode_t statusCode,
                   const char *fmt, ...) throw() _FANCYSOCK_WS_ATTR_FMT_PRINTF(5, 6);

  virtual ~WsCloseException() throw() {};

  /**
   * @return A WebSocket specific status code
   */
  inline wsStatusCode_t getStatusCode() const
  {
    return statusCode;
  }

  /**
   * @return An optional reason specifying the reason for the close.
   */
  inline const std::string& getReason() const
  {
    return reason;
  }
};

/**
 * WebSocket specific exception thrown if a protocol error occurs.
 */
class FANCYSOCK_API WsProtocolErrorException: public WsException {

public:
  inline WsProtocolErrorException(int facility, int priority,
                                  const char *fmt, ...) throw() _FANCYSOCK_WS_ATTR_FMT_PRINTF(4, 5):
    WsException(facility, priority)
  {
    va_list ap;
    va_start(ap, fmt);
    appendMessageV(fmt, ap);
    va_end(ap);
  }
};


/**
 * A Web Socket implementation as defined in
 * <a href="https://tools.ietf.org/html/rfc6455">RFC 6455</a>.
 */
class FANCYSOCK_API WsTransport: public virtual Fancy::Channel {

public:
  enum SocketState {
    Disconnected,
    Connecting,
    Connected,
    Closing,
    Closed
  };

  enum OpCode {
    Continuation = 0x0,
    Text         = 0x1,
    Binary       = 0x2,
    Close        = 0x8,
    Ping         = 0x9,
    Pong         = 0xA
  };

  enum DataType {
    TextData   = Text,
    BinaryData = Binary
  };

  enum FragmentationType {
    None    = 0x0,   // No fragmentation
    First   = 0x1,   // First message of a series of fragmented messages
    Next    = 0x2,   // Next message of a series of fragmented messages
    Last    = 0x3    // Last message of a series of fragmented messages
  };


  /**
   * Helper for logging
   */
  static const char* opCodeToStr(Fancy::Sock::WsTransport::OpCode opCode);

  /**
   * Helper for logging
   */
  static const char* dataTypeToStr(Fancy::Sock::WsTransport::DataType dataType);


  /**
   * @return eventual session
   *
   * Sessions are set in WsHttpDRequestHandler::handleRequest()
   * if a sessionManager was specified when c'ting WsHttpDRequestHandler
   */
  virtual Fancy::ObjPtr<HttpD::Session> getSession() const = 0;

  /**
   * Sets Sessions. Should only be used by WsHttpDRequestHandler::handleRequest().
   */
  //virtual void setSession(const Fancy::ObjPtr<HttpD::Session> &session) = 0;

  /**
   * @return The application data associated with this instance.
   */
  virtual Fancy::Object::ptr_t getData() const = 0;

  /**
   * Sets the application specific data.
   *
   * @param data Pointer to the new application specific data.
   */
  virtual void setData(const Fancy::Object::ptr_t & data) = 0;

  /**
   * @return The internally used socket-device.
   */
  virtual Fancy::ObjPtr<Device> getDevice() const = 0;

  /**
   * @return The maximum payload length, which incoming and outgoing frames may have.
   */
  virtual uint64_t getMaxPayloadLength() const = 0;

  /**
   * Sets the maximum payload length
   *
   * @param mayPayloadLength The new maximum payload length
   */
  virtual void setMaxPayloadLength(uint64_t maxPayloadLength) = 0;

  /**
   * @return Timeout (in msec), which is used whilst the WebSocket gets closed, for waiting
   *         for the reception of an appropriate close-acknowledgment sent through the remote
   *         host.
   */
  virtual int getCloseAckTimeout() const = 0;

  /**
   * Sets the timeout (in msec), which is used whilst the WebSocket gets closed for waiting
   * for the reception of an appropriate close-acknowledgment sent through the remote host.
   * If 'closeAckTimeout' is less-equal than zero, the default socket's read-timeout
   * (i.e. SO_RCVTIMEO) is used instead.
   *
   * @param timeout The new close timeout
   *
   * Default value: FANCYSOCK_WS_DEF_CLOSE_ACK_TIMEOUT
   */
  virtual void setCloseAckTimeout(int timeout) = 0;

  /**
   * @return The current ping interval
   */
  virtual int getPingInterval() const = 0;

  /**
   * @param interval The new ping interval (in msec) at which ping frames are sent while no
   *                 frames are received. By automatically sending ping frames, the WebSocket
   *                 won't get close if the client sends no messages for a longer time, but
   *                 the connection to the client is still good.
   *                 If 'pingInterval' is equal to zero, the half 'SO_RCVTIMEO' is used as
   *                 interval, except 'SO_RCVTIMEO' is infinite in which case 'pingInterval'
   *                 is also infinite.
   *                 If 'pingInterval' is less than zero, the automatic sending of ping
   *                 frames is disabled.
   */
  virtual void setPingInterval(int interval) = 0;

  /**
   * @return Whether ping frames are automatically answered with corresponding pong frames.
   */
  virtual bool isAutoPingReplyEnabled() const = 0;

  /**
   * @param enable Whether ping frames have to be answered automatically with corresponding
   *               pong frames.
   */
  virtual void setAutoPingReplyEnabled(bool enable) = 0;

  /**
   * @return Whether pong frames are handled transparently by the 'WsTransport' implementation
   *         or not.
   */
  virtual bool isIgnorePongEnabled() const = 0;

  /**
   * @param enable Whether to transparently handle pong frames. If false, received pong frames
   *               are forwarded to the application (i.e. returned by the 'beginFrameRd()'
   *               function).
   */
  virtual void setIgnorePongEnabled(bool enable) = 0;

  /**
   * @return The current socket state (@see SocketState)
   */
  virtual SocketState getSocketState() const = 0;

  /**
   * @return The name of the sub-protocol, which is used to communicate with the other host.
   *         If no sub-protocol is used, the function returns an empty string.
   */
  virtual const std::string& getSubProtocol() const = 0;

  /**
   * Begins reading a WebSocket frame by reading the next frame's header.
   * This function must be called before any of the 'xxxRd()' functions is used.
   *
   * @param timeout A timeout (in msec) for waiting for the next frame to be received. If
   *                'timeout' is less-equal than zero, the default socket's read timeout
   *                (i.e. SO_RCVTIMEO) is used instead.
   *
   * @return If the function times out, false is returned. Otherwise true is returned.
   *
   * Note: This function automatically ends reading the previous frame. This means, that all
   *       unread payload data of the previous frame is discarded by this function.
   *       If the function times out, no frame header is loaded at all. Thus, calling any of
   *       the 'xxxRd()' functions results in undefined behavior, if this function returns
   *       false.
   */
  virtual bool beginFrameRd(int timeout = 0) = 0;

  /**
   * Ends reading the current frame.
   * After this function has been called, none of the 'xxxRd()' functions must be called
   * until the next frame is loaded via an appropriate 'beginFrameRd()' call.
   *
   * @param timeout A timeout (in msec) for waiting for the next frame to be received. If
   *                'timeout' is less-equal than zero, the default socket's read timeout
   *                (i.e. SO_RCVTIMEO) is used instead.
   *
   * Note: If any payload data of the current frame hasn't been read so far, it gets
   *       discarded by this function.
   */
  virtual void endFrameRd(int timeout = 0) = 0;

  /**
   * @return The current frame's opcode.
   */
  virtual OpCode getOpCodeRd() const = 0;

  /**
   * @return The type of the data contained in the current frame's payload.
   */
  virtual DataType getDataTypeRd() const = 0;

  /**
   * @return The length of the data contained in the current frame's payload.
   */
  virtual uint64_t getPayloadLengthRd() const = 0;

  /**
   * @return The number of bytes, which still can be read from current frame's payload.
   */
  virtual uint64_t getPayloadAvailRd() const = 0;

  /**
   * Tries to read the specified amount of bytes of payload-data from the current frame
   *
   * @param  payload A buffer receiving the read data
   * @param  payloadLength The length of the buffer.
   *                       If 'payloadLength' exceeds the actual length of the payload, the
   *                       function only tries to read the amount of bytes actually available.
   * @param timeout A timeout (in msec) for reading the specified amount of bytes. If the
   *                function times out, the bytes read so far are returned or zero if no bytes
   *                were read at all. If 'timeout' is less-equal than zero, the default socket's
   *                read timeout (i.e. SO_RCVTIMEO) is used instead.
   *
   * @return The number of bytes actually read.
   */
  virtual int read(void *buf, int bufLen, int timeout) = 0;

  /**
   * Same as calling 'read(buf, bufLen, 0)'
   */
  virtual int read(void *buf, int bufLen) = 0;

  /**
   * @return The data-type used for the next outgoing frame
   *         (Default: FANCY_WEBSOCKET_DEF_DATA_TYPE_WR)
   */
  virtual DataType getDataTypeWr() const = 0;

  /**
   * Sets the data-type used  for the next outgoing frame
   * (Default: FANCY_WEBSOCKET_DEF_DATA_TYPE_WR)
   */
  virtual void setDataTypeWr(DataType dataType) = 0;

  /**
   * Creates a new frame with the given data as payload and sends it to the remote host.
   *
   * @param buf The frame's payload data.
   * @param bufLen The length of the payload data.
   * @param fragType The type of fragmentation to be used (@see FragmentationType)
   * @param dataType The data-type of the payload data.
   *
   * @return Number of bytes actually written. This is always equal to 'bufLen', as otherwise
   *         an exception is thrown.
   *
   * Note: If 'bufLen' exceeds the maximum payload length, an exception is thrown.
   */
  virtual int write(const void *buf, int bufLen, FragmentationType fragType, DataType dataType) = 0;

  /**
   * Same as calling 'write(buf, bufLen, fragType, getDataTypeWr())'
   */
  virtual int write(const void *buf, int bufLen, FragmentationType fragType) = 0;

  /**
   * Same as calling 'write(buf, bufLen, None, getDataTypeWr())'
   */
  virtual int write(const void *buf, int bufLen) = 0;

  /**
   * Creates a new frame with the given data as payload and sends it to the remote host.
   *
   * @param buf The frame's payload data.
   * @param bufLen The length of the payload data.
   * @param fragType The type of fragmentation to be used (@see FragmentationType)
   * @param dataType The data-type of the payload data.
   *
   * Note: If 'bufLen' exceeds the maximum payload length, an exception is thrown.
   *       If less than 'bufLen' bytes could be written, an exception is thrown.
   */
  virtual void writeAll(const void *buf, int bufLen, FragmentationType fragType,
                        DataType dataType) = 0;

  /**
   * Same as calling 'writeAll(buf, bufLen, fragType, getDataTypeWr())'
   */
  virtual void writeAll(const void *buf, int bufLen, FragmentationType fragType = None) = 0;

  /**
   * Sends a control-frame with an opcode 'Ping'
   *
   * @param payload Optional payload
   * @param payloadLen Size of the payload
   *
   * Note: The maximum payload-size of control-frames is 125 Bytes
   */
  virtual void sendPing(const void *payload = NULL, int payloadLen = 0) = 0;

  /**
   * Sends a control-frame with an opcode 'Pong'
   *
   * @param payload Optional payload
   * @param payloadLen Size of the payload
   *
   * Note: The maximum payload-size of control-frames is 125 Bytes
   */
  virtual void sendPong(const void *payload = NULL, int payloadLen = 0) = 0;

  /**
   * @see Fancy::Sock::Device::flush()
   */
  virtual void flush() = 0;

  /**
   * Performs a closing handshake and closes the underlying socket.
   *
   * @param statusCode A status code to be used for the closing handshake
   * @param reason An optional reason specifying why the connection gets closed.
   *
   * Note: If the closing handshake fails or the other side answers with a different status
   *       code than the specified one, an exception is thrown. This notwithstanding, the
   *       function always closes the underlying socket.
   */
  virtual void close(wsStatusCode_t statusCode, const std::string &reason = "") = 0;

  /**
   * Same as calling 'close(FANCYSOCK_WS_STATUS_NORMAL_CLOSE)'
   */
  virtual void close() = 0;

  /**
   * @see Fancy::Sock::Device::isOpen()
   */
  virtual bool isOpen() const = 0;

  /**
   * @see Fancy::Sock::Device::shutdown()
   *
   * Note: The shutdown-option is always set to 'ShutdownBoth'
   */
  virtual void shutdown(ShutdownOption of) = 0;

  /*
   * Sends a close frame.
   * This is similar to close(wsStatusCode_t, reason), but does not tear everything down,
   * to allow a smoother close and maybe avoid cornercase crashes.
   */
  virtual void sendCloseFrame(wsStatusCode_t statusCode, const std::string &reason) = 0;
};


/**
 * Server-side implementation of the WebSocket protocol
 */
class FANCYSOCK_API WsServerTransport: public virtual WsTransport {

public:
  typedef Fancy::ObjPtr<WsServerTransport> ptr_t;

  /**
   * Tries to upgrade an existing HTTP connection to a WebSocket connection.
   * If the upgrade fails, an exception is thrown.
   *
   * @param httpServerTransp A 'HttpServerTransport' instance (Must not be Null)
   * @param httpHeaders A map containing the HTTP request's headers (Must not be Null)
   * @param subProtocolIter An iterator pointing to the first item of a list of supported
   *                        sub-protocols. If 'subProtocolIter' is Null or the associated
   *                        list is empty, the server won't agree on any of the sub-protocols
   *                        requested by the client. However, if the client also specifies no
   *                        sub-protocol, the opening handshake may still succeed.
   * @param maxPayloadLen Maximum size of a frame's payload. If an incoming frame's payload
   *                      exceeds this value, the connection is terminated with a status code
   *                      of 'FANCYSOCK_WS_STATUS_TOO_BIG'. In case of outgoing frames, a
   *                      corresponding 'write()'-call, which exceeds this limit, causes an
   *                      appropriate exception to be thrown and no frame to be sent at all.
   * @param pingInterval Interval (in msec) at which ping frames are sent while no frames are
   *                     received. By automatically sending ping frames, the WebSocket won't get
   *                     close if the client sends no messages for a longer time, but the
   *                     connection to the client is still good.
   *                     If 'pingInterval' is equal to zero, the half 'SO_RCVTIMEO' is used as
   *                     interval, except 'SO_RCVTIMEO' is infinite in which case 'pingInterval'
   *                     is also infinite.
   *                     If 'pingInterval' is less than zero, the automatic sending of ping
   *                     frames is disabled.
   * @param autoPingReply Whether to send automatically a pong frame in response to a ping frame.
   *                      If false, the application is responsible for sending a pong frame as
   *                      soon as possible when a ping frame is received.
   * @param ignorePong If true, received pong frames are transparently handled by the
   *                   'WsServerTransport' instance. Otherwise received pong frames are
   *                   forwarded to the application (i.e. returned by the 'beginFrameRd()'
   *                   function).
   *
   * @return A WsServerTransport instance.
   */
  static ptr_t newInstance(const Fancy::ObjPtr<HttpServerTransport> &httpServerTransp,
                           const Fancy::ObjPtr<Map> &httpHeaders,
                           const Fancy::ObjPtr<Iterator> &subProtocolIter,
                           const Fancy::ObjPtr<HttpD::Session> &session,
                           uint64_t maxPayloadLen = FANCYSOCK_WS_DEF_MAX_PAYLOAD_LEN,
                           int pingInterval = 0,
                           bool autoPingReply = true,
                           bool ignorePong = true);

  virtual Fancy::ObjPtr<HttpD::Session> getSession() = 0;
};


}

}

#endif /* FANCYSOCKWSTRANSPORT_H_ */
